import copy
import importlib


def update_module_arguments(module, **kwargs):
    for k, v in kwargs.items():
        setattr(module, k, v)


class TPInfo(object):
    """
    A class to describe tensor parallelization information.

    Args:
        name (Tuple[str]): the name of parameter
        combined_qkv (bool): combined qkv or not
        parallel (bool): parallelizable param or not
        reverse (bool): reversed param or not
    """

    def __init__(
        self,
        *name,
        combined_qkv: bool = False,
        reverse: bool = False,
    ):
        self.name = name
        self.combined_qkv = combined_qkv
        self.reverse = reverse

    def __str__(self):
        return f"{self.__class__.__qualname__}({self.name})"

    def __repr__(self):
        return self.__str__()


Col = type("COLUMN", (TPInfo,), {"code": "Col"})
Row = type("ROW", (TPInfo,), {"code": "Row"})
Update = type("UPDATE", (TPInfo,), {"code": "Update"})


class TPMapping(object):
    __MAPPING__ = dict(
        Albert=[
            Col("query", "key", "value", "ffn"),
            Row("attention.dense", "ffn_output"),
            Update("num_attention_heads", "all_head_size"),
        ],
        Bart=[
            Col("q_proj", "k_proj", "v_proj", "fc1"),
            Row("out_proj", "fc2"),
            Update("embed_dim", "num_heads"),
        ],
        Bert=[
            Col("query", "key", "value", "intermediate.dense"),
            Row("output.dense"),
            Update("num_attention_heads", "all_head_size"),
        ],
        T5=[
            Col("q", "k", "v", "DenseReluDense.wi"),
            Row("o", "DenseReluDense.wo", "relative_attention_bias"),
            Update("d_model", "n_heads", "inner_dim"),
        ],
        GPT2=[
            Col("c_attn", reverse=True, combined_qkv=True),
            Col("c_fc", "q_attn", reverse=True),
            Row("c_proj", reverse=True),
            Update("embed_dim", "split_size", "num_heads"),
        ],
        GPTNeo=[
            Col("q_proj", "k_proj", "v_proj", "c_fc"),
            Row("out_proj", "c_proj"),
            Update("embed_dim", "num_heads"),
        ],
        GPTJ=[
            Col("q_proj", "k_proj", "v_proj", "fc_in"),
            Row("out_proj", "fc_out"),
            Update("embed_dim", "num_attention_heads"),
        ],
        Electra=[
            Col("query", "key", "value", "intermediate.dense"),
            Row("output.dense"),
            Update("num_attention_heads", "all_head_size"),
        ],
        Roberta=[
            Col("query", "key", "value", "intermediate.dense"),
            Row("output.dense"),
            Update("num_attention_heads", "all_head_size"),
        ],
    )

    def __init__(self):
        cache_tp_mapping = {}

        for cls_name, mapping in self.__MAPPING__.items():
            cls = self._load_class_by_model_name(cls_name)
            cache_tp_mapping[cls] = []

            for elem in mapping:
                for name in elem.name:
                    copy_elem = copy.deepcopy(elem)
                    copy_elem.name = name
                    cache_tp_mapping[cls].append(copy_elem)

        self.__MAPPING__ = {cls: {} for cls in cache_tp_mapping}
        # clear exist mapping rather than making new mapping dict

        for cls, mapping in cache_tp_mapping.items():
            for elem in mapping:
                if elem.code in self.__MAPPING__[cls]:
                    self.__MAPPING__[cls][elem.code].append(elem)
                else:
                    self.__MAPPING__[cls][elem.code] = [elem]

    @staticmethod
    def _load_class_by_model_name(model_name):
        """
        Load base class obj by class name
        Args:
            model_name (str): model name (e.g. Bert, GPT2, T5, ...)

        Returns:
            class: XXXPreTrainedModel
        """
        transformers = importlib.import_module("transformers")
        cls = getattr(transformers, f"{model_name}PreTrainedModel", None)
        if cls is None:
            cls = getattr(transformers, f"{model_name}PretrainedModel", None)
        assert cls is not None, f"Can not import the model named {cls}."
        return cls

    def get_mapping(self, model):
        """
        Get mapping by model obj

        Args:
            model (PreTrainedModel): model object (e.g. BertForSequenceClassification)

        Returns:
            dict: mapping by model
        """
        mapping_by_model = None
        for cls, mapping in self.__MAPPING__.items():
            if isinstance(model, cls):
                mapping_by_model = mapping

        assert mapping_by_model is not None, (
            f"Currently, {model.__class__.__qualname__} is not supported. "
            f"The current supported models are {list(self.__MAPPING__.keys())}"
        )
        return mapping_by_model

    def column_parallel_params(self, model):
        """
        Get list of column parallel param elements

        Args:
            model (PreTrainedModel): model obj

        Returns:
            List[COLUMN]: list of column parallel param elements
        """
        mapping = self.get_mapping(model)
        if mapping is not None:
            return mapping["Col"]

    def row_parallel_params(self, model):
        """
        Get list of row parallel param elements

        Args:
            model (PreTrainedModel): model obj

        Returns:
            List[ROW]: list of row parallel param elements
        """
        mapping = self.get_mapping(model)
        if mapping is not None:
            return mapping["Row"]

    def update_attrs(self, model):
        """
        Get list of update attribute elements

        Args:
            model (PreTrainedModel): model obj

        Returns:
            List[UPDATE]: list of update attribute elements
        """
        mapping = self.get_mapping(model)
        if mapping is not None:
            return mapping["Update"]

    def search(self, model, param_name):
        """
        Get element by parameter name

        Args:
            model (PreTrainedModel): model obj

        Returns:
            TPInfo: element by parameter name
        """
        mapping = self.get_mapping(model)
        count_contain_elem_in_param = 0
        param_split = param_name.split(".")
        first_check = []

        for elems in mapping.values():
            for elem in elems:
                if elem.name in param_name:
                    first_check.append(elem)

        for elem in first_check:
            elem_split = elem.name.split(".")
            for split in elem_split:
                if split in param_split:
                    count_contain_elem_in_param += 1
            if count_contain_elem_in_param == len(elem_split):
                return elem

        return None

    def is_combined_qkv_param(self, model, param_name):
        """
        Check whether the param is combined qkv or not

        Args:
            model (PreTrainedModel): model obj
            param_name (str): name of parameter

        Returns:
            bool: whether the param is combined qkv or not
        """
        elem = self.search(model, param_name)
        if elem is not None:
            return elem.combined_qkv

    def get_combined_qkv_degree(self, model, param_name, module):
        """
        Get combined qkv degree

        Args:
            model (PreTrainedModel): model obj
            param_name (str): name of parameter
            module (nn.Module): module that has `weight` parameter

        Returns:
            int: combined qkv degree
        """
        if self.is_combined_qkv_param(model, param_name) and hasattr(module, "weight"):
            bigger = max(module.weight.size(0), module.weight.size(1))
            smaller = min(module.weight.size(0), module.weight.size(1))
            return bigger // smaller
        return 1

    def is_reversed_param(self, model, param_name):
        """
        Check whether the parameter is reversed or not

        Args:
            model (PreTrainedModel): model obj
            param_name (str): name of parameter

        Returns:
            bool: whether the param is reversed or not
        """
        elem = self.search(model, param_name)
        if elem is not None:
            return elem.reverse

    def is_column_parallel(self, model, param_name):
        """
        Check whether the parameter is column parallelizable or not

        Args:
            model (PreTrainedModel): model obj
            param_name (str): name of parameter

        Returns:
            bool: whether the param is column parallelizable or not
        """
        elem = self.search(model, param_name)
        if elem is not None:
            return elem.code == "Col"

    def is_row_parallel(self, model, param_name):
        """
        Check whether the parameter is row parallelizable or not

        Args:
            model (PreTrainedModel): model obj
            param_name (str): name of parameter

        Returns:
            bool: whether the param is row parallelizable or not
        """
        elem = self.search(model, param_name)
        if elem is not None:
            return elem.code == "Row"



import importlib
import os

from oslo.pytorch.activation_checkpointing.checkpoint_engine import (
    ActivationCheckpointingEngine,
)

OPTIONS = {}


def initialize_activation_checkpointing(model, config, **kwargs):
    if "activation_checkpointing" in config:
        ac_config = config["activation_checkpointing"]

        if "enable" in ac_config and ac_config["enable"] is True:
            mpu = model.mpu if hasattr(model, "mpu") else None

            if mpu is None or mpu.get_tensor_parallel_world_size() <= 1:
                partitioned_checkpointing = ac_config.get(
                    "partitioned_checkpointing", False
                )
                contiguous_checkpointing = ac_config.get(
                    "contiguous_checkpointing", False
                )
                assert not partitioned_checkpointing and not contiguous_checkpointing, (
                    "``partitioned_checkpointing`` and ``contiguous_checkpointing`` can be used when you are used tensor model parallelism. "
                    "please set them as False if you don't use tensor model parallelism."
                )

            if ac_config.get("contiguous_checkpointing", False):
                assert ac_config.get("partitioned_checkpointing", False), (
                    "``contiguous_checkpointing`` is only available with ``partitioned_checkpointing``. "
                    "Set ``partitioned_checkpointing`` to true in your config."
                )

            engine = ActivationCheckpointingEngine(
                mpu=mpu,
                num_layers=model.config.num_hidden_layers,
                partitioned_checkpointing=ac_config.get(
                    "partitioned_checkpointing", False
                ),
                cpu_checkpointing=ac_config.get("cpu_checkpointing", False),
                contiguous_checkpointing=ac_config.get(
                    "contiguous_checkpointing", False
                ),
            )

            importlib.import_module(
                model.__module__
            ).torch.utils.checkpoint.checkpoint = engine.checkpoint

            model.gradient_checkpointing_enable()
            model = model.train()

    return model, config



import contextlib
import copy

import torch
from torch import _C
from torch.cuda import _lazy_call, device as device_ctx_manager


class CudaRNGStatesTracker(object):
    """
    Tracker for the cuda RNG states.

    Using the `add` method, a cuda rng state is initialized based on
    the input `seed` and is assigned to `name`. Later, by forking the
    rng state, we can perform operations and return to our starting
    cuda state.
    """

    def __init__(self, tracker_name="model-parallel-rng", mpu=None):
        self.mpu = mpu
        self.tracker_name = tracker_name
        self.states_ = {}
        self.seeds_ = set()

    def reset(self):
        """Set to the initial state (no tracker)."""
        self.states_ = {}
        self.seeds_ = set()

    def get_states(self):
        """Get rng states. Copy the dictionary so we have direct
        pointers to the states, not just a pointer to the dictionary."""
        return copy.copy(self.states_)

    def set_states(self, states):
        """Set the rng states. For efficiency purposes, we do not check
        the size of seed for compatibility."""
        self.states_ = states

    def add(self, name, seed):
        """Track the rng state."""
        # Check seed is not already used.
        if seed in self.seeds_:
            raise Exception("seed {} already exists".format(seed))
        self.seeds_.add(seed)
        # Check that state is not already defined.
        if name in self.states_:
            raise Exception("cuda rng state {} already exists".format(name))
        # Get the current rng state.
        orig_rng_state = torch.cuda.get_rng_state()
        # Set the new state and store it.
        torch.cuda.manual_seed(seed)
        self.states_[name] = torch.cuda.get_rng_state()
        # Reset rng state to what it was.
        self.set_cuda_rng_state(orig_rng_state)

    @contextlib.contextmanager
    def fork(self):
        """Fork the cuda rng state, perform operations, and exit with
        the original state."""
        # Check if we have added the state
        if self.tracker_name not in self.states_:
            raise Exception("cuda rng state {} is not added".format(self.tracker_name))
        # Store current rng state.
        orig_cuda_rng_state = torch.cuda.get_rng_state()
        # Set rng state to the desired one
        self.set_cuda_rng_state(self.states_[self.tracker_name])
        # Do the stuff we wanted to do.
        try:
            yield
        finally:
            # Update the current rng state for later use.
            self.states_[self.tracker_name] = torch.cuda.get_rng_state()
            # And set the state to the original state we started with.
            self.set_cuda_rng_state(orig_cuda_rng_state)

    @staticmethod
    def set_cuda_rng_state(new_state, device=-1):
        """
        Sets the random number generator state of the current GPU.

        Args:
            new_state (torch.ByteTensor): The desired state

        Notes:
            This function is adapted from PyTorch repo (torch.cuda.set_rng_state)
            with a single change: the input state is not cloned. Cloning caused
            major performance issues for +4 GPU cases.
        """
        if hasattr(_C, "_cuda_setRNGState") and callable(_C._cuda_setRNGState):
            # older PyTorch
            def cb():
                with device_ctx_manager(device):
                    _C._cuda_setRNGState(new_state)

        else:
            # newer PyTorch
            if device == -1:
                device = torch.device("cuda")
            elif isinstance(device, str):
                device = torch.device(device)
            elif isinstance(device, int):
                device = torch.device("cuda", device)

            def cb():
                idx = device.index
                if idx is None:
                    idx = torch.cuda.current_device()
                default_generator = torch.cuda.default_generators[idx]
                default_generator.set_state(new_state)

        _lazy_call(cb)



import json
import os
from typing import Any, Dict, Union

from oslo.pytorch.activation_checkpointing import initialize_activation_checkpointing
from oslo.pytorch.model_parallelism import initialize_model_parallelism


def _type(_type):
    return lambda key, val: {
        "check": isinstance(val, _type),
        "msg": f"``{key}: {val}`` is not a valid set. it must be type of {_type}",
    }


def _one_of_(*args):
    return lambda key, val: {
        "check": val in list(args),
        "msg": f"``{key}: {val}`` is not a valid set. it must be one of the {list(args)}",
    }


SUPPORTED_FEATURES = {
    "model_parallelism": {
        "enable": _type(bool),
        "tensor_parallel_size": _type(int),
    },
    "activation_checkpointing": {
        "enable": _type(bool),
        "partitioned_checkpointing": _type(bool),
        "contiguous_checkpointing": _type(bool),
        "cpu_checkpointing": _type(bool),
    },
}


def _config_check(arg, user):
    assert len(user) > 0, f"There are no arguments in dictionary."

    if isinstance(user, dict):
        for k in user:
            if isinstance(arg, dict):
                assert k in arg, (
                    f"An argument ``{k}`` is not available. "
                    f"We only support the arguments like {list(arg.keys())}."
                )
            else:
                raise Exception(
                    f"``{k}: {user[k]} is not a valid set. "
                    f"please check your configuration.``"
                )

            if isinstance(user[k], dict):
                _config_check(arg[k], user[k])
            else:
                assert not isinstance(arg[k], dict), (
                    f"``{k}: {user[k]} is not a valid set. "
                    f"please check your configuration.``"
                )

                check_result = arg[k](k, user[k])
                assert check_result["check"], check_result["msg"]
    else:
        raise TypeError("configuration must be type of <class 'dict'>")


def _sanity_check(
    model,
    config: Union[str, Dict[str, Any]],
):
    from transformers import PreTrainedModel

    assert isinstance(
        model, PreTrainedModel
    ), "An argument ``model`` must be the PyTorch Hugging Face Transformers model."

    assert isinstance(config, dict) or isinstance(
        config, str
    ), "An argument ``config`` must be the dictionary object or json path."

    if isinstance(config, str):
        if os.path.isfile(config):
            config = json.load(open(config, encoding="utf-8"))
        else:
            raise ValueError(
                f"Con not find the {config}. "
                f"Please double check your config file name."
            )

    _config_check(SUPPORTED_FEATURES, config)

    return model, config


def initialize(model, config: Union[str, Dict[str, Any]], **kwargs):
    """
    Initialize OSLO engine.

    Args:
        model (PreTrainedModel): The PyTorch Hugging Face Transformers model
        config (Union[str, os.PathLike, Dict[str, Any]]): dict object or json path
    """
    model, config = _sanity_check(model, config)
    model, config = initialize_model_parallelism(model, config, **kwargs)
    model, config = initialize_activation_checkpointing(model, config, **kwargs)
    return model



from functools import partial

from oslo.pytorch.model_parallelism.network.mpu import MPU
from oslo.pytorch.model_parallelism.tensor_parallel_enigne import (
    TensorDeparallelEngine,
    TensorParallelEngine,
)
from oslo.pytorch.model_parallelism.utils.extensions import (
    from_parallelized,
    resize_token_embeddings,
    save_parallelized,
)
from oslo.pytorch.model_parallelism.utils.mappings import TPMapping


def initialize_model_parallelism(model, config, **kwargs):
    if "model_parallelism" in config:
        mp_config = config["model_parallelism"]

        if "enable" in mp_config and mp_config["enable"] is True:
            tp_size = mp_config.get("tensor_parallel_size", 1)
            pp_size = mp_config.get("pipeline_parallel_size", 1)

            assert tp_size >= 1, "param `tensor_parallel_size` must be positive."
            assert pp_size >= 1, "param `pipeline_parallel_size` must be positive."
            assert (
                tp_size & (tp_size - 1) == 0
            ), "param `tensor_parallel_size` must be power of 2."

            if tp_size * pp_size > 1:
                mpu = MPU(tensor_parallel_size=tp_size, pipeline_parallel_size=pp_size)

                if tp_size > 1:
                    assert model.config.num_attention_heads % tp_size == 0, (
                        "``tensor_parallel_size`` must be divisible by ``num_attention_heads``. "
                        "Please check your model configuration."
                    )
                    assert model.config.hidden_size % tp_size == 0, (
                        "``tensor_parallel_size`` must be divisible by ``hidden_size``. "
                        "Please check your model configuration."
                    )
                    assert model.config.num_attention_heads >= tp_size, (
                        "``tensor_parallel_size`` must be same or greather than ``num_attention_heads``. "
                        "Please check your model configuration."
                    )
                    assert model.config.hidden_size >= tp_size, (
                        "``tensor_parallel_size`` must be same or greather than ``hidden_size``. "
                        "Please check your model configuration."
                    )

                    mapping = kwargs.pop("tp_mapping", TPMapping())
                    tensor_parallel_engine = TensorParallelEngine(model, mpu, mapping)
                    tensor_parallel_engine.parallelize()
                    setattr(
                        model,
                        "resize_token_embeddings",
                        partial(resize_token_embeddings, self=model),
                    )

                setattr(
                    model, "from_parallelized", partial(from_parallelized, self=model)
                )
                setattr(
                    model, "save_parallelized", partial(save_parallelized, self=model)
                )

    return model, config



from oslo.pytorch.model_parallelism.network.mpu import MPU
from oslo.pytorch.model_parallelism.utils.mappings import TPMapping
from oslo.pytorch.model_parallelism import TensorParallelEngine, TensorDeparallelEngine
from oslo.pytorch.activation_checkpointing import ActivationCheckpointingEngine



from oslo.pytorch.activation_checkpointing.checkpoint_function import CheckpointFunction
from oslo.pytorch.activation_checkpointing.checkpoint_partitioner import (
    CheckpointPartitioner,
)
from oslo.pytorch.activation_checkpointing.rng_state_tracker import CudaRNGStatesTracker


class ActivationCheckpointingEngine(object):
    def __init__(
        self,
        num_layers,
        partitioned_checkpointing,
        cpu_checkpointing,
        contiguous_checkpointing,
        mpu=None,
    ):
        rng_tracker = CudaRNGStatesTracker(mpu=mpu)
        partitioner = CheckpointPartitioner(
            mpu=mpu,
            num_layers=num_layers,
            partitioned_checkpointing=partitioned_checkpointing,
            contiguous_checkpointing=contiguous_checkpointing,
            cpu_checkpointing=cpu_checkpointing,
        )
        self.options = {
            "rng_tracker": rng_tracker,
            "partitioner": partitioner,
        }

    def checkpoint(self, function, *args):
        all_outputs = []
        CheckpointFunction.apply(function, self.options, all_outputs, *args)
        if len(all_outputs) == 1:
            return all_outputs
        else:
            return tuple(all_outputs)



# Copyright 2021 TUNiB Inc.
import math

import torch
import torch.nn.functional as F
from torch import nn


class VocabParallelEmbedding(nn.Embedding):
    """Embedding parallelized in the vocabulary dimension."""

    def _get_partitioned_vocab_range(self, world_size, rank):
        indices = [
            math.ceil(self.num_embeddings / world_size) for _ in range(world_size)
        ]

        indices[-1] = self.num_embeddings - sum(indices[:-1])
        indices.insert(0, 0)
        # 50257 => [0, 12565, 12565, 12565, 12652]

        for i in range(1, len(indices)):
            # cumulative summation
            # [0, 12565, 25130, 37595, 50257]
            indices[i] += indices[i - 1]

        return indices[rank], indices[rank + 1]

    def forward(self, inputs):
        world_size = self.mpu.get_tensor_parallel_world_size()

        if world_size > 1:
            indices = self._get_partitioned_vocab_range(
                world_size,
                rank=self.mpu.get_tensor_parallel_rank(),
            )

            vocab_start_index, vocab_end_index = indices
            input_mask = (inputs < vocab_start_index) | (inputs >= vocab_end_index)
            masked_input = inputs.clone() - vocab_start_index
            masked_input[input_mask] = 0
        else:
            masked_input = inputs

        output_parallel = F.embedding(
            masked_input,
            self.weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )

        if world_size > 1:
            output_parallel[input_mask, :] = 0.0

        out = self.mpu.reduce(output_parallel)
        return out


class ColumnParallelLinear(nn.Linear):
    """Linear layer with column parallelism."""

    def forward(self, inputs):
        inputs = self.mpu.broadcast(inputs)

        if self.reversed:
            outputs = torch.matmul(inputs, self.weight)
        else:
            outputs = torch.matmul(inputs, self.weight.t())

        if self.gather_output:
            outputs = self.mpu.gather(outputs).clone()

        if self.bias is not None:
            outputs += self.bias

        return outputs


class RowParallelLinear(nn.Linear):
    """Linear layer with row parallelism."""

    def forward(self, inputs):
        if self.reversed:
            outputs = torch.matmul(inputs, self.weight)
        else:
            outputs = torch.matmul(inputs, self.weight.t())

        outputs = self.mpu.reduce(outputs).clone()

        if self.bias is not None:
            outputs += self.bias

        return outputs



# Copyright 2021 TUNiB Inc.


from setuptools import find_packages, setup

VERSION = {}  # type: ignore

with open("oslo/__version__.py", "r") as version_file:
    exec(version_file.read(), VERSION)

with open("requirements.txt", "r") as requirements_file:
    INSTALL_REQUIRES = requirements_file.read().splitlines()

setup(
    name="oslo-core",
    description="OSLO: Open Source framework for Large-scale transformer Optimization",
    version=VERSION["version"],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/tunib-ai/oslo",
    author="TUNiB OSLO Team",
    author_email="contact@tunib.ai",
    install_requires=INSTALL_REQUIRES,
    packages=find_packages(
        include=["oslo", "oslo.*"],
        exclude=("tests", "tutorial", "docs"),
    ),
    python_requires=">=3.6.0",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    package_data={},
    dependency_links=[],
    include_package_data=True,
    zip_safe=False,
)



import copy
from typing import List

import torch
import torch.distributed as dist
from torch.nn import Embedding, Linear, Module

from oslo.pytorch.model_parallelism.utils.distributed import (
    ColumnParallelLinear,
    RowParallelLinear,
    VocabParallelEmbedding,
)
from oslo.pytorch.model_parallelism.utils.mappings import (
    TPMapping,
    update_module_arguments,
)


class TensorParallelEngine(object):
    def __init__(self, model, mpu, mapping=None):
        self.model = model
        self.mpu = mpu
        self.mapping = mapping if mapping is not None else TPMapping()
        self.device = torch.cuda.current_device()

    def _update_mp_arguments(self):
        for module in self.model.modules():
            for elem in self.mapping.update_attrs(self.model):
                if hasattr(module, elem.name):
                    world_size = self.mpu.get_tensor_parallel_world_size()
                    reduced_arg = getattr(module, elem.name) // world_size
                    setattr(module, elem.name, reduced_arg)

    @staticmethod
    def _deconstruct_combined_qkv(tensor, world_size):
        ranks = (len(tensor) + world_size - 1) // world_size
        tensor = [tensor[i * world_size : (i + 1) * world_size] for i in range(ranks)]
        tensor = list(map(lambda x: torch.cat([*x], dim=-1), zip(*tensor)))
        return tensor

    def _slice(
        self,
        module,
        reversed,
        fusion_degree,
        slice_bias,
        dim,
        to_gpu,
    ):
        dim = dim if not reversed else abs(dim - 1)
        world_size = self.mpu.get_tensor_parallel_world_size()
        gpu_index = self.mpu.get_tensor_parallel_rank()

        update_module_arguments(
            module=module,
            mpu=self.mpu,
            reversed=reversed,
            fusion_degree=fusion_degree,
            orig_module=copy.deepcopy(module.__class__),
            gather_output=False,
        )

        if hasattr(module, "weight") and module.weight is not None:
            if module.weight.dim() >= 1:
                weight = module.weight.chunk(fusion_degree * world_size, dim=dim)
                if fusion_degree > 1:
                    weight = self._deconstruct_combined_qkv(weight, world_size)
                module.weight.data = weight[gpu_index].contiguous()

            if to_gpu is True:
                module.weight.data = module.weight.to(self.device)

            update_module_arguments(
                module=module,
                in_features=module.weight.size()[0],
                out_features=module.weight.size()[1],
            )

        if hasattr(module, "bias") and module.bias is not None:
            if slice_bias is True and module.bias.dim() >= 1:

                bias = module.bias.chunk(fusion_degree * world_size, dim=0)
                if fusion_degree > 1:
                    bias = self._deconstruct_combined_qkv(bias, world_size)
                module.bias.data = bias[gpu_index].contiguous()

            if to_gpu is True:
                module.bias.data = module.bias.to(self.device)

        return module

    def _column_slice(
        self,
        module: Module,
        fusion_degree: int,
        reversed: bool,
        to_gpu: bool,
    ) -> Module:
        return self._slice(
            module=module,
            reversed=reversed,
            fusion_degree=fusion_degree,
            slice_bias=True,
            to_gpu=to_gpu,
            dim=0,
        )

    def _row_slice(
        self,
        module: Module,
        fusion_degree: int,
        reversed: bool,
        to_gpu: bool,
    ) -> List[Module]:
        return self._slice(
            module=module,
            reversed=reversed,
            fusion_degree=fusion_degree,
            slice_bias=False,
            to_gpu=to_gpu,
            dim=1,
        )

    def _parallelize_embedding(self):
        from transformers import Conv1D

        embedding = self.model.get_input_embeddings()
        chunked_weights = torch.chunk(
            embedding.weight, chunks=self.mpu.get_tensor_parallel_world_size(), dim=0
        )
        chunked_weight = chunked_weights[self.mpu.get_tensor_parallel_rank()]
        embedding.weight.data = chunked_weight.to(self.device)
        update_module_arguments(
            module=embedding,
            mpu=self.mpu,
            orig_module=copy.deepcopy(embedding.__class__),
        )

        if isinstance(embedding, Embedding):
            embedding.__class__ = VocabParallelEmbedding

        for name, module in self.model.named_modules():
            if (
                hasattr(module, "weight")
                and module.weight is embedding.weight
                and not isinstance(module, Embedding)
            ):
                update_module_arguments(
                    module=module,
                    mpu=self.mpu,
                    reversed=self.mapping.is_reversed_param(self.model, name),
                    fusion_degree=1,
                    orig_module=copy.deepcopy(module.__class__),
                    gather_output=True,
                )

                if isinstance(module, Linear) or isinstance(module, Conv1D):
                    module.__class__ = ColumnParallelLinear

    def _parallelize_modules(self):
        from transformers import Conv1D

        for param_name, module in self.model.named_modules():
            if self.mapping.is_column_parallel(self.model, param_name):
                self._column_slice(
                    module=module,
                    reversed=self.mapping.is_reversed_param(self.model, param_name),
                    fusion_degree=self.mapping.get_combined_qkv_degree(
                        self.model, param_name, module
                    ),
                    to_gpu=True,
                )
                if isinstance(module, Linear) or isinstance(module, Conv1D):
                    module.__class__ = ColumnParallelLinear

            elif self.mapping.is_row_parallel(self.model, param_name):
                self._row_slice(
                    module=module,
                    reversed=self.mapping.is_reversed_param(self.model, param_name),
                    fusion_degree=1,
                    to_gpu=True,
                )
                if isinstance(module, Linear) or isinstance(module, Conv1D):
                    module.__class__ = RowParallelLinear

    def _postprocess(self):
        for param in self.model.parameters():
            if not param.is_cuda:
                if torch.is_tensor(param):
                    param.data = param.to(self.device)

        for param in self.model.buffers():
            if not param.is_cuda:
                if torch.is_tensor(param):
                    param.data = param.to(self.device)

    def parallelize(self):
        self._update_mp_arguments()
        self._parallelize_embedding()
        self._parallelize_modules()
        self._postprocess()
        update_module_arguments(self.model, mpu=self.mpu)


class TensorDeparallelEngine(object):
    def __init__(self, model, mpu, mapping=None):
        self.model = model
        self.mpu = mpu
        self.mapping = mapping if mapping is not None else TPMapping()
        self.device = torch.cuda.current_device()

    def _update_mp_arguments(self):
        for module in self.model.modules():
            for elem in self.mapping.update_attrs(self.model):
                if hasattr(module, elem.name):
                    world_size = self.mpu.get_tensor_parallel_world_size()
                    reduced_arg = getattr(module, elem.name) * world_size
                    setattr(module, elem.name, reduced_arg)

    @torch.no_grad()
    def _reconstruct_combined_qkv(self, tensors, fusion_degree, dim):
        result_tensors = {i: [] for i in range(fusion_degree)}
        for tensor in tensors:
            chunks = tensor.chunk(fusion_degree, dim=dim)
            for i, chunk in enumerate(chunks):
                result_tensors[i].append(chunk)
        for key, val in result_tensors.items():
            result_tensors[key] = torch.cat(val, dim=dim)
        return torch.cat(list(result_tensors.values()), dim=dim)

    def _merge(
        self,
        module,
        reversed,
        fusion_degree,
        merge_bias,
        dim,
    ):
        dim = dim if not reversed else abs(dim - 1)
        world_size = self.mpu.get_tensor_parallel_world_size()
        update_module_arguments(module=module, mpu=None)

        if hasattr(module, "weight") and module.weight is not None:
            if module.weight.dim() >= 1:
                if not module.weight.is_contiguous():
                    module.weight.data = module.weight.contiguous()
                if not module.weight.is_cuda:
                    module.weight.data = module.weight.to(self.device)

                tensor_list = [
                    torch.zeros_like(module.weight) for _ in range(world_size)
                ]

                dist.all_gather(
                    tensor_list,
                    module.weight,
                    group=self.mpu.get_tensor_parallel_group(),
                )

                if fusion_degree > 1:
                    output = self._reconstruct_combined_qkv(
                        tensor_list, fusion_degree, dim
                    )
                else:
                    output = torch.cat(tensor_list, dim=dim)

                module.weight.data = output
                update_module_arguments(
                    module=module,
                    in_features=module.weight.size()[0],
                    out_features=module.weight.size()[1],
                    nx=module.weight.size()[0],
                    nf=module.weight.size()[1],
                )

            module.weight.data = module.weight.cpu()

        if hasattr(module, "bias") and module.bias is not None:
            if merge_bias is True and module.bias.dim() >= 1:
                if not module.bias.is_contiguous():
                    module.bias.data = module.bias.contiguous()
                if not module.bias.is_cuda:
                    module.bias.data = module.bias.to(self.device)

                tensor_list = [torch.zeros_like(module.bias) for _ in range(world_size)]

                dist.all_gather(
                    tensor_list,
                    module.bias,
                    group=self.mpu.get_tensor_parallel_group(),
                )

                if fusion_degree > 1:
                    output = self._reconstruct_combined_qkv(
                        tensor_list, fusion_degree, 0
                    )
                else:
                    output = torch.cat(tensor_list, dim=0)

                module.bias.data = output

            module.bias.data = module.bias.cpu()

        return module

    def _column_merge(
        self,
        module: Module,
        fusion_degree: int,
        reversed: bool,
    ) -> Module:
        from transformers import Conv1D

        merged_module = self._merge(
            module=module,
            reversed=reversed,
            fusion_degree=fusion_degree,
            merge_bias=True,
            dim=0,
        )

        if hasattr(module, "orig_module"):
            module.__class__ = module.orig_module
        else:
            module.__class__ = Conv1D if reversed else Linear

        return merged_module

    def _row_merge(
        self,
        module: Module,
        fusion_degree: int,
        reversed: bool,
    ) -> List[Module]:
        from transformers import Conv1D

        merged_module = self._merge(
            module=module,
            reversed=reversed,
            fusion_degree=fusion_degree,
            merge_bias=False,
            dim=1,
        )

        if hasattr(module, "orig_module"):
            module.__class__ = module.orig_module
        else:
            module.__class__ = Conv1D if reversed else Linear

        return merged_module

    def _deparallelize_embedding(self):
        from transformers import Conv1D

        embedding = self.model.get_input_embeddings()
        if not embedding.weight.is_cuda:
            embedding.weight.data = embedding.weight.cuda()

        gathered_weight = self.mpu._gather(embedding.weight, dim=0)
        embedding.weight.data = gathered_weight.cpu()

        update_module_arguments(
            module=embedding,
            mpu=None,
            num_embeddings=embedding.weight.size()[0],
        )

        if hasattr(embedding, "orig_module"):
            embedding.__class__ = embedding.orig_module
        else:
            embedding.__class__ = Embedding

        for module in self.model.modules():
            if (
                hasattr(module, "weight")
                and module.weight is embedding.weight
                and not isinstance(module, Embedding)
            ):
                update_module_arguments(
                    module=module,
                    mpu=None,
                    out_features=embedding.num_embeddings,
                )

                if hasattr(module, "orig_module"):
                    module.__class__ = module.orig_module
                else:
                    if hasattr(module, "reversed"):
                        module.__class__ = Conv1D if module.reversed else Linear
                    else:
                        module.__class__ = Linear

                module.weight.data = module.weight.cpu()

    def _deparallelize_modules(self):
        for param_name, module in self.model.named_modules():
            if self.mapping.is_column_parallel(self.model, param_name):
                self._column_merge(
                    module=module,
                    fusion_degree=module.fusion_degree,
                    reversed=module.reversed,
                )
            elif self.mapping.is_row_parallel(self.model, param_name):
                self._row_merge(
                    module=module,
                    fusion_degree=module.fusion_degree,
                    reversed=module.reversed,
                )

    def _postprocess(self):
        for param in self.model.parameters():
            if param.is_cuda:
                if torch.is_tensor(param):
                    param.data = param.cpu()

        for param in self.model.buffers():
            if param.is_cuda:
                if torch.is_tensor(param):
                    param.data = param.cpu()

    def deparallelize(self):
        self._update_mp_arguments()
        self._deparallelize_embedding()
        self._deparallelize_modules()
        self._postprocess()
        update_module_arguments(self.model, mpu=None)



# Copyright 2021 TUNiB Inc.

import os
from typing import List

import torch
import torch.distributed as dist
from torch import Size, Tensor
from torch.autograd import Function


class MPU(object):
    """
    MPU: Model Parallel Unit

    MPU is key concept of Oslo framework and is inspired by Megatron-LM.
    The main difference with Megatron-LM is that each model has an their mpu.

    We can combine several models later. For example, in the case of Electra,
    there is a generator model and a discriminator model. To parallelize all of them,
    each model must be parallelized in a different process group,
    so the mpu must be maintained in the model level, not the global state.

    Notes:
        Let's say we have a total of 16 GPUs denoted g0 ... g15 and we use 2 GPUs to parallelize the model tensor,
        and 4 GPUs to parallelize the model pipeline. The present method will create 8 model-parallel group,
        4 pipeline parallel groups and 8 data parallel groups as:

        - width: 4 pipeline parallel group
            [g0, g4, g8, g12], [g1, g5, g9, g13], [g2, g6, g10, g14], [g3, g7, g11, g15]
        - height: 8 tensor parallel group
            [g0, g1], [g2, g3], [g4, g5], [g6, g7], [g8, g9], [g10, g11], [g12, g13], [g14, g15]
        - depth: 8 data parallel group
            [g0, g2], [g1, g3], [g4, g6], [g5, g7], [g8, g10], [g9, g11], [g12, g14], [g13, g15]

                        [g02, g06, g10, g14]
                      /  |              /  |
                     [g00, g04, g08, g12]  |
                     |   |             |   |
        3D parallel  |  [g03, g07, g11, g15]
                     |  /              |  /
                     [g01, g05, g09, g13]

                     +-----+  +----------+  +----------+  +----------+  +----------+  +-----+
              model  | g00 |  |   g00    |  |   g04    |  |   g08    |  |   g12    |  | g12 |
        data         +-----+  +----------+  +----------+  +----------+  +----------+  +-----+  ===> forward
              model  | g01 |  |   g01    |  |   g05    |  |   g09    |  |   g13    |  | g13 |
                     +-----+  +----------+  +----------+  +----------+  +----------+  +-----+
                    embedding   pipeline      pipeline      pipeline      pipeline   embedding

                     +-----+  +----------+  +----------+  +----------+  +----------+  +-----+
              model  | g02 |  |   g02    |  |   g06    |  |   g10    |  |   g14    |  | g14 |
        data         +-----+  +----------+  +----------+  +----------+  +----------+  +-----+  ===> forward
              model  | g03 |  |   g03    |  |   g07    |  |   g11    |  |   g15    |  | g15 |
                     +-----+  +----------+  +----------+  +----------+  +----------+  +-----+
                    embedding   pipeline      pipeline      pipeline      pipeline   embedding

    References:
        Original MPU implementation of Megatron-LM. We refactored all the code to be pythonic.
        https://github.com/NVIDIA/Megatron-LM/blob/main/megatron/mpu/initialize.py

    """

    # process group
    _data_parallel_group = None
    _model_parallel_group = None
    _tensor_parallel_group = None
    _pipeline_parallel_group = None

    # world size
    _data_parallel_world_size = None
    _model_parallel_world_size = None
    _tensor_parallel_world_size = None
    _pipeline_parallel_world_size = None

    # single rank
    _data_parallel_rank = None
    _model_parallel_rank = None
    _tensor_parallel_rank = None
    _pipeline_parallel_rank = None

    # global ranks
    _pipeline_parallel_global_ranks_per_group = None
    _pipeline_parallel_global_ranks_across_all_groups = []

    def __init__(self, tensor_parallel_size: int, pipeline_parallel_size: int) -> None:
        """
        Initialize MPU object.
        All the process groups are initialized when this method is invoked.

        Args:
            tensor_parallel_size (int): model parallel world size
            pipeline_parallel_size (int): pipeline parallel world size
            master_port (int): master port to initialize global process group
        """

        if not dist.is_initialized():
            self.initialize_distributed()

        current_rank = dist.get_rank()
        global_world_size = dist.get_world_size()

        assert (
            global_world_size >= tensor_parallel_size
        ), "param `tensor_parallel_size` must be smaller than global world size."

        assert (
            global_world_size >= pipeline_parallel_size
        ), "param `pipeline_model_parallel_size` must be smaller than global world size."

        total_model_parallel_size = tensor_parallel_size * pipeline_parallel_size

        assert (
            global_world_size % total_model_parallel_size == 0
        ), "global world sizes must be divisible by model parallel world sizes (tp * pp)"

        num_tensor_parallel_groups = global_world_size // tensor_parallel_size
        num_pipeline_parallel_groups = global_world_size // pipeline_parallel_size

        # 1. initialize data parallel group
        all_data_parallel_group_ranks = self._initialize_data_parallel_group(
            current_rank=current_rank,
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            num_pipeline_parallel_groups=num_pipeline_parallel_groups,
        )

        self._initialize_model_parallel_groups(
            current_rank=current_rank,
            all_data_parallel_group_ranks=all_data_parallel_group_ranks,
        )

        # 3. initialize tensor model parallel group
        self._initialize_tensor_parallel_group(
            current_rank=current_rank,
            tensor_parallel_size=tensor_parallel_size,
            num_tensor_parallel_groups=num_tensor_parallel_groups,
        )

        # 4. initialize pipeline model parallel group
        self._initialize_pipeline_parallel_group(
            current_rank=current_rank,
            global_world_size=global_world_size,
            num_pipeline_parallel_groups=num_pipeline_parallel_groups,
        )

        # 5. create distributed functions
        functions = self._initialize_functions()
        self._broadcast_fn = functions["broadcast"]
        self._reduce_fn = functions["reduce"]
        self._scatter_fn = functions["scatter"]
        self._gather_fn = functions["gather"]

    # Initialization

    @staticmethod
    def initialize_distributed():
        """Initialize torch.distributed and mpu."""
        rank = int(os.getenv("RANK", 0))
        world_size = int(os.getenv("WORLD_SIZE", 1))

        if not dist.is_initialized():
            device_count = torch.cuda.device_count()

            if device_count > 0:
                device = rank % device_count
                torch.cuda.set_device(device)

            torch.distributed.init_process_group(
                backend="nccl",
                world_size=world_size,
                rank=rank,
            )

    def _initialize_data_parallel_group(
        self,
        current_rank: int,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        num_pipeline_parallel_groups: int,
    ) -> List[List[int]]:
        """
        Initialize data parallel group

        Args:
            current_rank (int): current rank
            tensor_parallel_size (int): tensor model parallel world size
            pipeline_parallel_size (int): pipeline model parallel world size
            num_pipeline_parallel_groups (int): the number of pipeline model parallel groups

        Returns:
            List[List[int]]: all_data_parallel_group_ranks
        """
        assert (
            self._data_parallel_group is None
        ), "data parallel group is already initialized."

        all_data_parallel_group_ranks = []
        for i in range(pipeline_parallel_size):
            start_rank = i * num_pipeline_parallel_groups
            end_rank = (i + 1) * num_pipeline_parallel_groups

            for j in range(tensor_parallel_size):
                ranks = list(range(start_rank + j, end_rank, tensor_parallel_size))
                all_data_parallel_group_ranks.append(ranks)
                group = dist.new_group(ranks)
                if current_rank in ranks:
                    self._data_parallel_group = group

        return all_data_parallel_group_ranks

    def _initialize_model_parallel_groups(
        self,
        current_rank: int,
        all_data_parallel_group_ranks: List[List[int]],
    ):
        # Build the model-parallel groups.
        assert (
            self._model_parallel_group is None
        ), "model parallel group is already initialized"
        for i in range(self.get_data_parallel_world_size()):
            ranks = [
                data_parallel_group_ranks[i]
                for data_parallel_group_ranks in all_data_parallel_group_ranks
            ]
            group = torch.distributed.new_group(ranks)
            if current_rank in ranks:
                self._model_parallel_group = group

    def _initialize_tensor_parallel_group(
        self,
        current_rank: int,
        tensor_parallel_size: int,
        num_tensor_parallel_groups: int,
    ) -> None:
        """
        Initialize tensor model parallel group

        Args:
            current_rank (int): current rank
            tensor_parallel_size (int): tensor model parallel world size
            num_tensor_parallel_groups (int): the number of tensor model parallel groups
        """
        assert (
            self._tensor_parallel_group is None
        ), "tensor model parallel group is already initialized."

        for i in range(num_tensor_parallel_groups):
            start_rank = i * tensor_parallel_size
            end_rank = (i + 1) * tensor_parallel_size

            ranks = list(range(start_rank, end_rank))
            group = dist.new_group(ranks)

            if current_rank in ranks:
                self._tensor_parallel_group = group

    def _initialize_pipeline_parallel_group(
        self,
        current_rank: int,
        global_world_size: int,
        num_pipeline_parallel_groups: int,
    ) -> None:
        """
        Initialize pipeline model parallel group

        Args:
            current_rank (int): current rank
            global_world_size (int): global world size
            num_pipeline_parallel_groups (int): the number of tensor model parallel groups
        """
        assert (
            self._pipeline_parallel_group is None
        ), "pipeline model parallel group is already initialized."

        for i in range(num_pipeline_parallel_groups):
            ranks = list(range(i, global_world_size, num_pipeline_parallel_groups))
            group = dist.new_group(ranks)

            if current_rank in ranks:
                self._pipeline_parallel_group = group
                self._pipeline_model_parallel_global_ranks_per_group = ranks

            if len(ranks) > 1:
                embedding_ranks = [ranks[0], ranks[-1]]
            else:
                embedding_ranks = ranks

            group = torch.distributed.new_group(embedding_ranks)

            if current_rank in embedding_ranks:
                self._embedding_tied_group = group
            if current_rank in ranks:
                self._embedding_tied_global_ranks = embedding_ranks

            self._pipeline_parallel_global_ranks_across_all_groups.append(ranks)

    def model_parallel_is_initialized(self) -> bool:
        """
        Check if model parallel groups are initialized.

        Returns:
            bool: whether model parallel groups are initialized
        """
        if (
            self._tensor_parallel_group is None
            or self._pipeline_parallel_group is None
            or self._data_parallel_group is None
        ):
            return False
        return True

    # Process group

    def get_data_parallel_group(self):
        """
        Get the data parallel group.

        Returns:
            ProcessGroup: data parallel group
        """
        assert (
            self._data_parallel_group is not None
        ), "data parallel group is not initialized."

        return self._data_parallel_group

    def get_model_parallel_group(self):
        """
        Get the model parallel group.

        Returns:
            ProcessGroup: model parallel group
        """
        assert (
            self._model_parallel_group is not None
        ), "model parallel group is not initialized."

        return self._model_parallel_group

    def get_tensor_parallel_group(self):
        """
        Get the tensor model parallel group.

        Returns:
            ProcessGroup: tensor model parallel group
        """
        assert (
            self._tensor_parallel_group is not None
        ), "tensor model parallel group is not initialized."

        return self._tensor_parallel_group

    def get_pipeline_parallel_group(self):
        """
        Get the pipeline model parallel group.

        Returns:
            ProcessGroup: pipeline model parallel group
        """
        assert (
            self._pipeline_parallel_group is not None
        ), "pipeline model parallel group is not initialized."

        return self._pipeline_parallel_group

    # World Size

    def get_data_parallel_world_size(self) -> int:
        """
        Get the data parallel world size

        Returns:
            int: data parallel world size
        """
        if self._data_parallel_world_size is not None:
            return self._data_parallel_world_size

        return dist.get_world_size(self.get_data_parallel_group())

    def get_model_parallel_world_size(self):
        """
        Get the model parallel world size

        Returns:
            int: model parallel world size
        """
        if self._model_parallel_world_size is not None:
            return self._model_parallel_world_size

        return dist.get_world_size(self.get_model_parallel_group())

    def get_tensor_parallel_world_size(self) -> int:
        """
        Get the tensor model parallel world size

        Returns:
            int: tensor model parallel world size
        """
        if self._tensor_parallel_world_size is not None:
            return self._tensor_parallel_world_size

        return dist.get_world_size(self.get_tensor_parallel_group())

    def get_pipeline_parallel_world_size(self) -> int:
        """
        Get the pipeline model parallel world size

        Returns:
            int: pipeline model parallel world size
        """
        if self._pipeline_parallel_world_size is not None:
            return self._pipeline_parallel_world_size

        return dist.get_world_size(self.get_pipeline_parallel_group())

    # Single Rank

    def get_data_parallel_rank(self) -> int:
        """
        Get the data parallel rank

        Returns:
            int: data parallel rank
        """
        if self._data_parallel_rank is not None:
            return self._data_parallel_rank

        return dist.get_rank(self.get_data_parallel_group())

    def get_model_parallel_rank(self) -> int:
        """
        Get the model parallel rank

        Returns:
            int: model parallel rank
        """
        if self._model_parallel_rank is not None:
            return self._model_parallel_rank

        return dist.get_rank(self.get_model_parallel_group())

    def get_tensor_parallel_rank(self) -> int:
        """
        Get the tensor model parallel rank

        Returns:
            int: tensor model parallel world size
        """
        if self._tensor_parallel_rank is not None:
            return self._tensor_parallel_rank

        return dist.get_rank(self.get_tensor_parallel_group())

    def get_pipeline_parallel_rank(self) -> int:
        """
        Get pipeline model parallel rank

        Returns:
            int: pipeline model parallel rank
        """
        if self._pipeline_parallel_rank is not None:
            return self._pipeline_parallel_rank

        return dist.get_rank(self.get_pipeline_parallel_group())

    def get_tensor_parallel_src_rank(self) -> int:
        """
        Compute the global rank corresponding to the first local rank in the tensor model parallel group.

        Returns:
            int: tensor model parallel source rank
        """
        global_rank = dist.get_rank()
        local_world_size = self.get_tensor_parallel_world_size()
        return (global_rank // local_world_size) * local_world_size

    def is_pipeline_first_stage(self) -> bool:
        """
        Return `True` if in the first pipeline model parallel stage, `False` otherwise

        Returns:
            bool: whether current pipeline model parallel stage is first
        """
        return self.get_pipeline_parallel_rank() == 0

    def is_pipeline_last_stage(self) -> bool:
        """
        Return `True` if in the last pipeline model parallel stage, `False` otherwise

        Returns:
            bool: whether current pipeline model parallel stage is last
        """
        return self.get_pipeline_parallel_rank() == (
            self.get_pipeline_parallel_world_size() - 1
        )

    # Global Ranks

    def get_pipeline_parallel_global_ranks_across_all_groups(self):
        """
        Get all the pipeline model parallel ranks across all groups

        Returns:
            List[List[int]]: all the parallel ranks across all devices
        """
        return self._pipeline_parallel_global_ranks_across_all_groups

    def get_pipeline_model_parallel_global_ranks_per_group(self):
        """
        Get all the pipeline model parallel ranks per single group

        Returns:
            List[int]: all the parallel ranks
        """
        return self._pipeline_model_parallel_global_ranks_per_group

    # Utils functions

    def destroy_model_parallel(self) -> None:
        """
        Destroy all the model parallel groups
        """

        self._tensor_parallel_group = None
        self._pipeline_parallel_group = None
        self._data_parallel_group = None

    def _broadcast(self, inputs: Tensor) -> Tensor:
        """
        Pass the input to the tensor model parallel region.

        Args:
            inputs (Tensor): input tensor

        Returns:
            Tensor: broadcast tensor
        """

        if not inputs.is_cuda:
            inputs = inputs.cuda()

        return inputs

    def _reduce(self, inputs: Tensor):
        """
        All-reduce the input tensor across tensor model parallel group.

        Args:
            inputs (Tensor): input tensor

        Returns:
            Tensor: all-reduced tensor
        """
        if self.get_tensor_parallel_world_size() == 1:
            return inputs

        dist.all_reduce(
            tensor=inputs,
            group=self.get_tensor_parallel_group(),
        )

        return inputs

    def _scatter(self, inputs: Tensor, dim: int = -1) -> Tensor:
        """
        Split the tensor along given dimension and keep the corresponding slice.

        Args:
            inputs (Tensor): input tensor
            dim (int): scatter dimension

        Returns:
            Tensor: scattered tensor
        """
        world_size = self.get_tensor_parallel_world_size()

        if world_size == 1:
            return inputs

        inputs_list = torch.chunk(input=inputs, chunks=world_size, dim=dim)
        # In megatron-lm, `torch.split` is used,
        # but `torch.chunk` is correct to scatter odd-sized tensors.

        rank = self.get_tensor_parallel_rank()
        outputs = inputs_list[rank].contiguous()
        return outputs

    def _gather(self, inputs: Tensor, dim: int = -1) -> Tensor:
        """
        Gather tensors and concatenate along the given dimension
        We support different size tensor all-gather.

        Args:
            inputs (Tensor): input tensor
            dim (int): gather dimension

        Returns:
            Tensor: gathered tensor
        """

        world_size = self.get_tensor_parallel_world_size()

        if world_size == 1:
            return inputs

        input_size = list(inputs.size())

        size_tensor = torch.tensor(
            input_size[dim], device=inputs.device, dtype=torch.long
        )
        sizes_list = [torch.empty_like(size_tensor) for _ in range(world_size)]
        dist.all_gather(sizes_list, size_tensor, group=self.get_tensor_parallel_group())

        sizes_list = [_.item() for _ in sizes_list]
        max_size = max(sizes_list)
        pad_size = max_size - input_size[dim]

        if pad_size > 0:
            pad_tensor_size = list(inputs.size())
            pad_tensor_size[dim] = pad_size
            pad = torch.zeros(
                *pad_tensor_size,
                device=inputs.device,
                dtype=inputs.dtype,
            )
            inputs = torch.cat([inputs, pad], dim=dim)

        tensor_list = [
            torch.empty(
                *inputs.size(),
                device=inputs.device,
                dtype=inputs.dtype,
            )
            for _ in range(world_size)
        ]

        if torch.is_tensor(inputs):
            if not inputs.is_contiguous():
                inputs = inputs.contiguous()

        dist.all_gather(tensor_list, inputs, group=self.get_tensor_parallel_group())

        # remove padding ;)
        for i, orig_size in enumerate(sizes_list):
            tensor_list[i] = torch.split(tensor_list[i], orig_size, dim=dim)[0]

        return torch.cat(tensor_list, dim=dim)

    def broadcast(self, inputs: Tensor) -> Tensor:
        """
        Pass the input to the tensor model parallel region.

        Args:
            inputs (Tensor):

        Returns:
            Tensor: broadcast tensor
        """

        if self._enable_grad(inputs):
            outputs = self._broadcast_fn.apply(inputs)
        else:
            outputs = self._broadcast(inputs)
        return outputs

    def reduce(self, inputs: Tensor) -> Tensor:
        """
        All-reduce the input tensor across tensor model parallel group.

        Args:
            inputs (Tensor): input tensor

        Returns:
            Tensor: all-reduced tensor
        """

        if self._enable_grad(inputs):
            outputs = self._reduce_fn.apply(inputs)
        else:
            outputs = self._reduce(inputs)
        return outputs

    def scatter(self, inputs: Tensor) -> Tensor:
        """
        Split the tensor along its last dimension and keep the corresponding slice.

        Args:
            inputs (Tensor): input tensor

        Returns:
            Tensor: scattered tensor
        """

        if self._enable_grad(inputs):
            outputs = self._scatter_fn.apply(inputs)
        else:
            outputs = self._scatter(inputs)
        return outputs

    def gather(self, inputs: Tensor) -> Tensor:
        """
        Gather tensors and concatenate along the last dimension

        Args:
            inputs (Tensor): input tensor

        Returns:
            Tensor: gathered tensor
        """

        if self._enable_grad(inputs):
            outputs = self._gather_fn.apply(inputs)
        else:
            outputs = self._gather(inputs)
        return outputs

    @staticmethod
    def _enable_grad(inputs: Tensor) -> bool:
        """
        Check current tensor is enabled to pass gradient.

        Args:
            inputs (Tensor): input tensor

        Returns:
            bool: whether gradient can be passed or not
        """
        return torch.is_grad_enabled() and inputs.requires_grad

    def _initialize_functions(self):
        class Broadcast(Function):
            @staticmethod
            def forward(ctx, inputs):
                return self._broadcast(inputs)

            @staticmethod
            def backward(ctx, inputs):
                return self._reduce(inputs)

        class Reduce(Function):
            @staticmethod
            def forward(ctx, inputs):
                return self._reduce(inputs)

            @staticmethod
            def backward(ctx, inputs):
                return self._broadcast(inputs)

        class Scatter(Function):
            @staticmethod
            def forward(ctx, inputs):
                return self._scatter(inputs)

            @staticmethod
            def backward(ctx, inputs):
                return self._gather(inputs)

        class Gather(Function):
            @staticmethod
            def forward(ctx, inputs):
                return self._gather(inputs)

            @staticmethod
            def backward(ctx, inputs):
                return self._scatter(inputs)

        return {
            "broadcast": Broadcast,
            "reduce": Reduce,
            "scatter": Scatter,
            "gather": Gather,
        }



import torch
from deepspeed.runtime.utils import move_to_device
from functools import partial
from oslo.pytorch.activation_checkpointing.utils import (
    extract_tensors,
    copy_to_device,
    is_activation_to_checkpoint,
    detach,
)


class CheckpointFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, run_function, options, all_outputs, *args):
        def save_args_for_backward(*all_args):
            tensor_args, non_tensor_args, tensor_flags = extract_tensors(all_args)
            ctx._saved_tensors = tensor_args
            ctx.non_tensor_args = non_tensor_args
            ctx.tensor_flags = tensor_flags

        # 1. save some variables for backward.
        ctx.rng_tracker = options.get("rng_tracker", None)
        ctx.partitioner = options.get("partitioner", None)
        ctx.run_function = run_function

        # 2. partitioning or moving to cpu if user want.
        if ctx.partitioner.partitioned_checkpointing:
            inputs = ctx.partitioner.make_partitioned_activations(args)
        elif ctx.partitioner.cpu_checkpointing:
            inputs = copy_to_device(
                args,
                torch.device("cpu"),
                partial(
                    is_activation_to_checkpoint,
                    world_size=ctx.partitioner.mp_size,
                ),
            )

        inputs_cuda = copy_to_device(
            args,
            torch.cuda.current_device(),
            partial(
                is_activation_to_checkpoint,
                world_size=ctx.partitioner.mp_size,
            ),
        )

        # 3. save rng states to mimic in backward.
        ctx.fwd_cpu_rng_state = torch.get_rng_state()
        ctx.fwd_cuda_rng_state = torch.cuda.get_rng_state()
        ctx.fwd_cuda_rng_state_tracker = ctx.rng_tracker.get_states()

        # 4. run forward without gradients
        with torch.no_grad():
            outputs = run_function(*inputs_cuda)
        del inputs_cuda

        # 5. save inputs of partition_checkpoint or cpu_checkpoint for backward stage.
        if ctx.partitioner.partitioned_checkpointing:
            new_args = ctx.partitioner.get_partitioned_activations_for_backward(
                args, inputs
            )
            assert len(new_args) % 2 == 0, (
                f"save_for_backward must call with even number of args. "
                f"but currently called with odd number of args, num_args={len(new_args)}"
            )
            save_args_for_backward(*new_args)
        elif ctx.partitioner.cpu_checkpointing:
            new_args = ctx.partitioner.get_cpu_activations_for_backward(args, inputs)
            save_args_for_backward(*new_args)
        else:
            save_args_for_backward(*args)

        # 6. make the non-fp tensors non-differential
        if torch.is_tensor(outputs):
            non_grad_outputs = [outputs] if not outputs.is_floating_point() else []
        else:
            non_grad_outputs = [
                o for o in outputs if torch.is_tensor(o) and not o.is_floating_point()
            ]
        ctx.mark_non_differentiable(*non_grad_outputs)

        # 7. return the results of forward stage.
        if torch.is_tensor(outputs):
            all_outputs += [outputs]
            return outputs
        else:
            all_outputs += outputs
            tensor_outputs, non_tensor_outputs, _ = extract_tensors(all_objects=outputs)
            return tuple(outputs)

    @staticmethod
    def backward(ctx, *grads):
        # 1. frees up all the pointers if user want contiguous checkpointing.
        if ctx.partitioner.contiguous_checkpointing:
            ctx.partitioner.contiguous_data_buffers = []
            ctx.partitioner.contiguous_size_buffers = []
            ctx.partitioner.data_offsets = []
            ctx.partitioner.size_offsets = []

        if not torch.autograd._is_checkpoint_valid():
            raise RuntimeError(
                "Checkpointing is not compatible with .grad(), "
                "please use .backward() if possible"
            )

        # 2. gathering or moving to gpu if user want and detach all the tensors.
        if ctx.partitioner.partitioned_checkpointing:
            inputs = ctx.partitioner.gather_partitioned_activations(
                ctx._saved_tensors,
                device=torch.cuda.current_device()
                if ctx.partitioner.cpu_checkpointing
                else None,
            )
            detached_inputs = detach(inputs)
            ctx.tensor_flags = ctx.tensor_flags[0::2]
            ctx.non_tensor_args = ctx.non_tensor_args[0::2]
        elif ctx.partitioner.cpu_checkpointing:
            inputs = move_to_device(
                ctx._saved_tensors,
                torch.cuda.current_device(),
                partial(
                    is_activation_to_checkpoint,
                    world_size=ctx.partitioner.mp_size,
                ),
            )
            detached_inputs = detach(inputs)
        else:
            inputs = ctx._saved_tensors
            detached_inputs = detach(inputs)

        # 3. merge_tensors if tensors are partitioned.
        detached_inputs = ctx.partitioner.merge_tensors(
            tensor_objects=detached_inputs,
            non_tensor_objects=ctx.non_tensor_args,
            tensor_flags=ctx.tensor_flags,
        )

        # 4. restore rng states to mimic in backward.
        bwd_cpu_rng_state = torch.get_rng_state()
        bwd_cuda_rng_state = torch.cuda.get_rng_state()
        bwd_cuda_rng_state_tracker = ctx.rng_tracker.get_states()

        # 5. set the states to what it used to be before the forward pass.
        torch.set_rng_state(ctx.fwd_cpu_rng_state)
        ctx.rng_tracker.set_cuda_rng_state(ctx.fwd_cuda_rng_state)
        ctx.rng_tracker.set_states(ctx.fwd_cuda_rng_state_tracker)

        # 6. run forward with gradients
        with torch.enable_grad():
            outputs = ctx.run_function(*detached_inputs)

        # 7. set the states back to what it was at the start of this function.
        torch.set_rng_state(bwd_cpu_rng_state)
        ctx.rng_tracker.set_cuda_rng_state(bwd_cuda_rng_state)
        ctx.rng_tracker.set_states(bwd_cuda_rng_state_tracker)

        if isinstance(outputs, torch.Tensor):
            outputs = (outputs,)

        # 8. filter out non tensor outputs.
        outputs, _, _ = extract_tensors(all_objects=outputs)

        # 9. run backward stage.
        output_tensors = []
        grad_tensors = []
        for out, grad in zip(outputs, grads):
            if out.requires_grad:
                output_tensors.append(out)
                grad_tensors.append(grad)
        torch.autograd.backward(output_tensors, grad_tensors)

        # 10. force clear our stashed tensors to prevent a memory leak in certain scenarios.
        ctx._saved_tensors = None
        ctx.non_tensor_args = None
        ctx.tensor_flags = None

        # 11. return gradients
        ret_list = [None, None, None]
        for inp in detached_inputs:
            if torch.is_tensor(inp):
                ret_list.append(inp.grad)
            else:
                ret_list.append(None)

        return tuple(ret_list)



import mmap

import torch
import torch.distributed as dist
from oslo.pytorch.activation_checkpointing.utils import is_activation_to_checkpoint


class CheckpointPartitioner(object):
    def __init__(
        self,
        mpu,
        num_layers,
        cpu_checkpointing,
        partitioned_checkpointing,
        contiguous_checkpointing,
    ):
        self.mp_size = mpu.get_tensor_parallel_world_size() if mpu is not None else 1
        self.mp_rank = mpu.get_tensor_parallel_rank() if mpu is not None else 0
        self.mp_group = mpu.get_tensor_parallel_group() if mpu is not None else None

        self.num_layers = num_layers
        self.cpu_checkpointing = cpu_checkpointing
        self.partitioned_checkpointing = partitioned_checkpointing
        self.contiguous_checkpointing = contiguous_checkpointing

        self.contiguous_data_buffers = []
        self.contiguous_size_buffers = []
        self.data_offsets = []
        self.size_offsets = []

    def get_partition_start(self, item):
        size = item.numel()
        partition_size = size / self.mp_size
        start = partition_size * self.mp_rank
        return int(start)

    def get_partition_size(self, item):
        size = item.numel()
        assert size % self.mp_size == 0, (
            "Doesn't handle if partition activation "
            "if item is not divisible by mp size."
        )
        partition_size = size / self.mp_size
        return int(partition_size)

    def make_partitioned_activations(self, args):
        inputs = []
        num_non_ckpt_tensors = 0

        for arg_index, item in enumerate(args):
            if not is_activation_to_checkpoint(item, self.mp_size):
                inputs.append(item)
                num_non_ckpt_tensors += 1
                continue

            i = arg_index - num_non_ckpt_tensors
            partition_size = self.get_partition_size(item)

            partition = (
                item.detach()
                .contiguous()
                .view(-1)
                .narrow(0, self.get_partition_start(item), partition_size)
                .clone()
            )

            buffer_device = (
                torch.device("cpu") if self.cpu_checkpointing else partition.device
            )

            if self.contiguous_checkpointing:
                if i >= len(self.contiguous_data_buffers):
                    tensor_list = [
                        torch.tensor(()).new_empty(
                            [partition_size],
                            dtype=partition.dtype,
                            device=buffer_device,
                        )
                        for _ in range(self.num_layers)
                    ]
                    self.contiguous_data_buffers.append(tensor_list)
                    self.data_offsets.append(0)

                elif self.contiguous_data_buffers[i] is None:
                    tensor_list = [
                        torch.tensor(()).new_empty(
                            [partition_size],
                            dtype=partition.dtype,
                            device=buffer_device,
                        )
                        for _ in range(self.num_layers)
                    ]
                    self.contiguous_data_buffers[i] = tensor_list
                    self.data_offsets[i] = 0

                # Because the 'new_empty' returns uninitialized pages,
                # the pages need to be populated during the cudaMemcpy time
                # which increases the data copy time. To avoid this, we
                # pre-populate these pages by simply writing 0 ahead of
                # the actual cudaMemcpy operation time. Due to the
                # previously launched GPU kernels, there is a small
                # window of time here for CPUs to populate pages asynchronously.
                self.contiguous_data_buffers[i][self.data_offsets[i]].data[
                    range(
                        0,
                        self.contiguous_data_buffers[i][
                            self.data_offsets[i]
                        ].data.shape[0],
                        int(
                            mmap.PAGESIZE
                            / self.contiguous_data_buffers[i][
                                self.data_offsets[i]
                            ].data.element_size()
                        ),
                    )
                ] = 0

                contiguous_partition = self.contiguous_data_buffers[i][
                    self.data_offsets[i]
                ].data.copy_(partition.data)
                self.data_offsets[i] = self.data_offsets[i] + 1
                inputs.append(contiguous_partition)
            else:
                partition = partition.cpu() if self.cpu_checkpointing else partition
                inputs.append(partition)

        return inputs

    def gather_partitioned_activations(self, tensors, device=None):
        assert (
            len(tensors) % 2 == 0
        ), f"Expected even count of tensors, instead got {len(tensors)}"

        inputs = []
        num_args = int(len(tensors) / 2)

        for i in range(num_args):
            item = tensors[2 * i]
            size = tensors[2 * i + 1]

            if not is_activation_to_checkpoint(item, self.mp_size):
                inputs.append(item)
                continue

            partition_size = item.numel()
            tensor_size = partition_size * self.mp_size

            if device is not None:
                flat_tensor = torch.zeros(
                    [tensor_size],
                    dtype=item.dtype,
                    device=device,
                )
            else:
                flat_tensor = torch.zeros(
                    [tensor_size],
                    dtype=item.dtype,
                    device=item.device,
                )

            partitions = []
            for i in range(self.mp_size):
                part_i = flat_tensor.narrow(0, partition_size * i, partition_size)
                if i == self.mp_rank:
                    part_i.copy_(item)
                partitions.append(part_i)

            if self.mp_group is not None:
                dist.all_gather(
                    partitions,
                    partitions[self.mp_rank],
                    group=self.mp_group,
                )

            input_tensor = flat_tensor.view(list(size.numpy()))
            item.data = input_tensor.data
            inputs.append(item)

        return tuple(inputs)

    @staticmethod
    def merge_tensors(tensor_objects, non_tensor_objects, tensor_flags):
        tensor_idx = 0
        non_tensor_idx = 0
        merged_objects = []

        for is_tensor in tensor_flags:
            if is_tensor:
                merged_objects.append(tensor_objects[tensor_idx])
                tensor_idx += 1
            else:
                merged_objects.append(non_tensor_objects[non_tensor_idx])
                non_tensor_idx += 1

        return tuple(merged_objects)

    def get_partitioned_activations_for_backward(self, args, inputs):
        new_args = []
        num_non_ckpt_tensors = 0

        for arg_index, (arg, inp) in enumerate(zip(args, inputs)):
            size = torch.tensor(arg.size()) if torch.is_tensor(arg) else None

            if not is_activation_to_checkpoint(arg, self.mp_size):
                new_args.append(arg)
                new_args.append(size)
                num_non_ckpt_tensors += 1
                continue

            arg.data = inp.data
            new_args.append(arg)
            i = arg_index - num_non_ckpt_tensors

            if self.contiguous_checkpointing:
                numel = size.numel()
                if i >= len(self.contiguous_size_buffers):
                    tmp = torch.tensor(())
                    self.contiguous_size_buffers.append(
                        tmp.new_empty(
                            [numel * self.num_layers],
                            dtype=size.dtype,
                            device=size.device,
                        )
                    )
                    self.size_offsets.append(0)
                elif self.contiguous_size_buffers[i] is None:
                    tmp = torch.tensor(())
                    self.contiguous_size_buffers[i] = tmp.new_empty(
                        [numel * self.num_layers],
                        dtype=size.dtype,
                        device=size.device,
                    )
                    self.size_offsets[i] = 0

                contiguous_size = (
                    self.contiguous_size_buffers[i]
                    .narrow(0, self.size_offsets[i], numel)
                    .data.copy_(size.data)
                )
                contiguous_size = contiguous_size.view_as(size)
                self.size_offsets[i] = self.size_offsets[i] + numel
                new_args.append(contiguous_size)
            else:
                new_args.append(size)

        return new_args

    def get_cpu_activations_for_backward(self, args, inputs):
        new_args = []
        for i, (arg, inp) in enumerate(zip(args, inputs)):
            if not is_activation_to_checkpoint(arg, self.mp_size):
                new_args.append(arg)
                continue

            arg.data = inp.data
            new_args.append(arg)

        return new_args



import os
from collections import Callable
from logging import getLogger
from typing import Optional, Union

import torch
import torch.distributed as dist

from oslo.pytorch.model_parallelism.tensor_parallel_enigne import (
    TensorDeparallelEngine,
    TensorParallelEngine,
)

PARALLELIZED_WEIGHTS_NAME = "pytorch_model_tp_0_pp_0.bin"

logger = getLogger(__name__)


def from_parallelized(parallelized_model_path, **kwargs):
    self = kwargs.pop("self")

    file_names = {
        os.path.join(
            parallelized_model_path,
            PARALLELIZED_WEIGHTS_NAME.replace("tp_0", f"tp_{tp}").replace(
                "pp_0", f"pp_{pp}"
            ),
        )
        for tp in range(self.mpu.get_tensor_parallel_world_size())
        for pp in range(self.mpu.get_pipeline_parallel_world_size())
    }

    if os.path.isdir(parallelized_model_path):
        if all(os.path.isfile(file_name) for file_name in file_names):
            state_dict = torch.load(
                os.path.join(
                    parallelized_model_path,
                    PARALLELIZED_WEIGHTS_NAME.replace(
                        "tp_0",
                        f"tp_{self.mpu.get_tensor_parallel_rank()}",
                    ).replace(
                        "pp_0",
                        f"pp_{self.mpu.get_pipeline_parallel_rank()}",
                    ),
                )
            )

            if self._keys_to_ignore_on_save is not None:
                state_dict = {
                    k: v
                    for k, v in state_dict.items()
                    if k not in self._keys_to_ignore_on_save
                }

            self.load_state_dict(state_dict, strict=False)

        else:
            raise FileNotFoundError(
                f"all the {file_names} are necessary. "
                f"but some of them do not exist. Please check your checkpoint files."
            )
    else:
        raise NotADirectoryError(
            f"directory named {parallelized_model_path} is not valid. "
        )

    return self


@torch.no_grad()
def save_parallelized(
    save_directory: Union[str, os.PathLike],
    save_config: bool = True,
    state_dict: Optional[dict] = None,
    save_function: Callable = torch.save,
    merge_checkpoints: bool = False,
    **kwargs,
):
    from transformers.modeling_utils import get_parameter_dtype, unwrap_model

    self = kwargs.pop("self")
    mapping = kwargs.pop("tp_mapping", None)

    if (
        self.mpu.get_tensor_parallel_world_size() == 1
        and self.mpu.get_pipeline_parallel_world_size() == 1
    ):
        if dist.get_rank() == 0:
            self.save_pretrained(
                save_directory=save_directory,
                save_config=save_config,
                state_dict=state_dict,
                save_function=save_function,
                **kwargs,
            )
        dist.barrier()
        return None

    if merge_checkpoints:
        model_to_save = self.__class__(self.config).eval()
        tensor_parallel_engine = TensorParallelEngine(model_to_save, self.mpu, mapping)
        tensor_parallel_engine.parallelize()

        if state_dict is None:
            state_dict = unwrap_model(self).state_dict()

        model_to_save.load_state_dict(state_dict)

        if self.mpu.get_tensor_parallel_world_size() > 1:
            tensor_deparallel_engine = TensorDeparallelEngine(
                model_to_save, model_to_save.mpu, mapping
            )
            tensor_deparallel_engine.deparallelize()

        if dist.get_rank() == 0:
            model_to_save.save_pretrained(
                save_directory=save_directory,
                save_config=save_config,
                save_function=save_function,
                **kwargs,
            )

        del model_to_save

        dist.barrier()
        return None

    if os.path.isfile(save_directory):
        logger.error(
            f"Provided path ({save_directory}) should be a directory, not a file"
        )
        return

    os.makedirs(save_directory, exist_ok=True)

    # Only save the model itself if we are using distributed training
    model_to_save = unwrap_model(self)

    # save the string version of dtype to the config, e.g. convert torch.float32 => "float32"
    # we currently don't use this setting automatically, but may start to use with v5
    dtype = get_parameter_dtype(model_to_save)
    model_to_save.config.torch_dtype = str(dtype).split(".")[1]

    # Attach architecture to the config
    model_to_save.config.architectures = [model_to_save.__class__.__name__]

    # Save the config
    if save_config:
        model_to_save.config.save_pretrained(save_directory)

    # Save the model
    if state_dict is None:
        state_dict = model_to_save.state_dict()

    # Handle the case where some state_dict keys shouldn't be saved
    if self._keys_to_ignore_on_save is not None:
        state_dict = {
            k: v for k, v in state_dict.items() if k not in self._keys_to_ignore_on_save
        }

    # If we save using the predefined names, we can load using `from_pretrained`
    weights_name = PARALLELIZED_WEIGHTS_NAME
    weights_name = weights_name.replace(
        "tp_0", f"tp_{self.mpu.get_tensor_parallel_rank()}"
    )
    weights_name = weights_name.replace(
        "pp_0", f"pp_{self.mpu.get_pipeline_parallel_rank()}"
    )

    output_model_file = os.path.join(save_directory, weights_name)

    if self.mpu.get_data_parallel_world_size() > 1:
        if self.mpu.get_data_parallel_rank() == 0:
            save_function(state_dict, output_model_file)
    else:
        save_function(state_dict, output_model_file)

    dist.barrier()
    logger.info(f"Model weights saved in {output_model_file}")


def resize_token_embeddings(new_num_tokens: Optional[int] = None, **kwargs):
    raise RuntimeError(
        "you can't use ``model.resize_token_embeddings()`` when you are using tensor model parallelism.\n"
        "please resize token embedding size before tensor parallelization."
    )



import torch


def is_activation_to_checkpoint(item, world_size):
    """
    Check is activation to checkpoint or not

    Args:
        item (Any): any object
        world_size (int): model parallel world size

    Returns:
        bool: is activation to checkpoint or not
    """
    return (
        torch.is_tensor(item)
        and item.is_floating_point()
        and item.numel() >= world_size
        and item.requires_grad
    )


def copy_to_device(item, device, criterion):
    """
    Copy item to specific device when criterion is satisfied.

    Args:
        item (Any): any object
        device (device): torch device
        criterion (function): criterion function

    Returns:
        Union[Tensor, List, Tuple, Dict]: items that is copied to specific device
    """
    if torch.is_tensor(item) and criterion(item):
        return item.to(device)
    elif isinstance(item, list):
        return [copy_to_device(i, device, criterion) for i in item]
    elif isinstance(item, tuple):
        return tuple([copy_to_device(i, device, criterion) for i in item])
    elif isinstance(item, dict):
        return {k: copy_to_device(v, device, criterion) for k, v in item.items()}


def detach(args, device=None):
    """
    Detach all the tensors in the tuple of data.

    Args:
        args (Tuple[Any]): tuple of data
        device (torch.device): device object

    Returns:
        Tuple[Any]: tuple of detached data.
    """
    if isinstance(args, tuple):
        output = []
        for arg in args:
            if not isinstance(arg, torch.Tensor):
                output.append(arg)
                continue

            requires_grad = arg.requires_grad

            if device is not None:
                x = arg.to(device=device)
            else:
                x = arg

            x = x.detach()
            x.requires_grad = requires_grad
            output.append(x)

        return tuple(output)
    else:
        raise RuntimeError(
            "Only tuple of tensors is supported. Got Unsupported input type: ",
            type(args).__name__,
        )


def extract_tensors(all_objects):
    """
    Separate objects in list/tuple into tensors and non-tensors and create a mapping to enable re-aggregation.
    The order of tensors and non-tensors is preserved in their respective output groups.

    Args:
        all_objects (list/tuple): Objects containing tensors and non-tensors to be split.

    Returns:
        tuple: Containing tensors, non-tensors, and bools of whether each position in original list/tuple was a tensor.
    """
    tensor_objects = [v for v in all_objects if torch.is_tensor(v)]
    non_tensor_objects = [v for v in all_objects if not torch.is_tensor(v)]
    tensor_flags = [torch.is_tensor(v) for v in all_objects]

    if type(all_objects) is tuple:
        return tuple(tensor_objects), tuple(non_tensor_objects), tuple(tensor_flags)

    return tensor_objects, non_tensor_objects, tensor_flags
